@@ -17,58 +17,67 @@
  */
 package org.apache.hadoop.fs.viewfs;
 
+import java.io.File;
 import java.io.FileNotFoundException;
+import java.io.FileOutputStream;
 import java.io.IOException;
 import java.net.URI;
-import java.util.ArrayList;
 import java.security.PrivilegedExceptionAction;
 import java.util.Arrays;
+import java.util.Collection;
 import java.util.List;
-
+import java.util.ArrayList;
+import java.util.Map;
+import java.util.Map.Entry;
 import java.util.Random;
+
 import org.apache.hadoop.conf.Configuration;
 import org.apache.hadoop.fs.BlockLocation;
-import org.apache.hadoop.fs.LocalFileSystem;
-import org.apache.hadoop.fs.TestFileUtil;
-import org.apache.hadoop.fs.Trash;
-import org.apache.hadoop.fs.contract.ContractTestUtils;
+import org.apache.hadoop.fs.BlockStoragePolicySpi;
+import org.apache.hadoop.fs.ContentSummary;
+import org.apache.hadoop.fs.FSDataOutputStream;
+import org.apache.hadoop.fs.FileStatus;
 import org.apache.hadoop.fs.FileSystem;
 import org.apache.hadoop.fs.FileSystemTestHelper;
-import static org.apache.hadoop.fs.FileSystemTestHelper.*;
-import org.apache.hadoop.fs.permission.AclEntry;
-import static org.apache.hadoop.fs.viewfs.Constants.PERMISSION_555;
-import static org.apache.hadoop.fs.viewfs.Constants.CONFIG_VIEWFS_ENABLE_INNER_CACHE;
-import static org.apache.hadoop.fs.viewfs.Constants.CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT;
-import static org.apache.hadoop.fs.FileSystem.TRASH_PREFIX;
-import static org.apache.hadoop.test.GenericTestUtils.*;
-import static org.junit.Assert.*;
-
-import org.apache.hadoop.fs.FileStatus;
 import org.apache.hadoop.fs.FsConstants;
+import org.apache.hadoop.fs.FsStatus;
+import org.apache.hadoop.fs.LocalFileSystem;
 import org.apache.hadoop.fs.LocatedFileStatus;
 import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.fs.RemoteIterator;
+import org.apache.hadoop.fs.TestFileUtil;
+import org.apache.hadoop.fs.Trash;
 import org.apache.hadoop.fs.UnsupportedFileSystemException;
+import org.apache.hadoop.fs.contract.ContractTestUtils;
 import org.apache.hadoop.fs.permission.AclEntry;
 import org.apache.hadoop.fs.permission.AclStatus;
 import org.apache.hadoop.fs.permission.AclUtil;
-import org.apache.hadoop.fs.RemoteIterator;
 import org.apache.hadoop.fs.permission.FsAction;
 import org.apache.hadoop.fs.permission.FsPermission;
-import org.apache.hadoop.fs.viewfs.ConfigUtil;
-import org.apache.hadoop.fs.viewfs.ViewFileSystem;
 import org.apache.hadoop.fs.viewfs.ViewFileSystem.MountPoint;
 import org.apache.hadoop.security.AccessControlException;
 import org.apache.hadoop.security.Credentials;
 import org.apache.hadoop.security.UserGroupInformation;
 import org.apache.hadoop.security.token.Token;
 import org.apache.hadoop.test.GenericTestUtils;
+import org.junit.Assume;
+import org.junit.Rule;
+import org.junit.rules.TemporaryFolder;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
+import static org.apache.hadoop.fs.FileSystemTestHelper.*;
+import static org.apache.hadoop.fs.viewfs.Constants.CONFIG_VIEWFS_ENABLE_INNER_CACHE;
+import static org.apache.hadoop.fs.viewfs.Constants.PERMISSION_555;
+
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
 import org.junit.Test;
 
+import static org.apache.hadoop.test.GenericTestUtils.assertExceptionContains;
 import static org.hamcrest.CoreMatchers.containsString;
+import static org.junit.Assert.*;
+
 
 /**
  * <p>
@@ -96,6 +105,8 @@ abstract public class ViewFileSystemBaseTest {
   Path targetTestRoot;
   Configuration conf;
   final FileSystemTestHelper fileSystemTestHelper;
+  private static final Logger LOG =
+      LoggerFactory.getLogger(ViewFileSystemBaseTest.class);
 
   public ViewFileSystemBaseTest() {
       this.fileSystemTestHelper = createFileSystemHelper();
@@ -105,6 +116,9 @@ abstract public class ViewFileSystemBaseTest {
     return new FileSystemTestHelper();
   }
 
+  @Rule
+  public TemporaryFolder temporaryFolder = new TemporaryFolder();
+
   @Before
   public void setUp() throws Exception {
     initializeTargetTestRoot();
@@ -158,6 +172,10 @@ abstract public class ViewFileSystemBaseTest {
   public void testGetMountPoints() {
     ViewFileSystem viewfs = (ViewFileSystem) fsView;
     MountPoint[] mountPoints = viewfs.getMountPoints();
+    for (MountPoint mountPoint : mountPoints) {
+      LOG.info("MountPoint: " + mountPoint.getMountedOnPath() + " => "
+          + mountPoint.getTargetFileSystemURIs()[0]);
+    }
     Assert.assertEquals(getExpectedMountPoints(), mountPoints.length); 
   }
   
@@ -249,7 +267,7 @@ abstract public class ViewFileSystemBaseTest {
         fsTarget.isFile(new Path(targetTestRoot,"user/foo")));
     
     // Delete the created file
-    Assert.assertTrue("Delete should suceed",
+    Assert.assertTrue("Delete should succeed",
         fsView.delete(new Path("/user/foo"), false));
     Assert.assertFalse("File should not exist after delete",
         fsView.exists(new Path("/user/foo")));
@@ -264,7 +282,7 @@ abstract public class ViewFileSystemBaseTest {
         fsTarget.isFile(new Path(targetTestRoot,"dir2/foo")));
     
     // Delete the created file
-    Assert.assertTrue("Delete should suceed",
+    Assert.assertTrue("Delete should succeed",
         fsView.delete(new Path("/internalDir/linkToDir2/foo"), false));
     Assert.assertFalse("File should not exist after delete",
         fsView.exists(new Path("/internalDir/linkToDir2/foo")));
@@ -954,6 +972,67 @@ abstract public class ViewFileSystemBaseTest {
     fsView.deleteSnapshot(new Path("/internalDir"), "snap1");
   }
 
+  @Test(expected = AccessControlException.class)
+  public void testInternalSetStoragePolicy() throws IOException {
+    fsView.setStoragePolicy(new Path("/internalDir"), "HOT");
+  }
+
+  @Test(expected = AccessControlException.class)
+  public void testInternalUnsetStoragePolicy() throws IOException {
+    fsView.unsetStoragePolicy(new Path("/internalDir"));
+  }
+
+  @Test(expected = AccessControlException.class)
+  public void testInternalSatisfyStoragePolicy() throws IOException {
+    fsView.satisfyStoragePolicy(new Path("/internalDir"));
+  }
+
+  @Test(expected = NotInMountpointException.class)
+  public void testInternalgetStoragePolicy() throws IOException {
+    fsView.getStoragePolicy(new Path("/internalDir"));
+  }
+
+  @Test
+  public void testInternalGetAllStoragePolicies() throws IOException {
+    Collection<? extends BlockStoragePolicySpi> policies =
+        fsView.getAllStoragePolicies();
+    for (FileSystem fs : fsView.getChildFileSystems()) {
+      try {
+        for (BlockStoragePolicySpi s : fs.getAllStoragePolicies()) {
+          assertTrue("Missing policy: " + s, policies.contains(s));
+        }
+      } catch (UnsupportedOperationException e) {
+        // ignore
+      }
+    }
+  }
+
+  @Test
+  public void testConfLinkSlash() throws Exception {
+    String clusterName = "ClusterX";
+    URI viewFsUri = new URI(FsConstants.VIEWFS_SCHEME, clusterName,
+        "/", null, null);
+
+    Configuration newConf = new Configuration();
+    ConfigUtil.addLink(newConf, clusterName, "/",
+        new Path(targetTestRoot, "/").toUri());
+
+    String mtPrefix = Constants.CONFIG_VIEWFS_PREFIX + "." + clusterName + ".";
+    try {
+      FileSystem.get(viewFsUri, newConf);
+      fail("ViewFileSystem should error out on mount table entry: "
+          + mtPrefix + Constants.CONFIG_VIEWFS_LINK + "." + "/");
+    } catch (Exception e) {
+      if (e instanceof UnsupportedFileSystemException) {
+        String msg = " Use " + Constants.CONFIG_VIEWFS_LINK_MERGE_SLASH +
+            " instead";
+        assertThat(e.getMessage(), containsString(msg));
+      } else {
+        fail("Unexpected exception: " + e.getMessage());
+      }
+    }
+  }
+
   @Test
   public void testTrashRoot() throws IOException {
 
@@ -1022,174 +1101,70 @@ abstract public class ViewFileSystemBaseTest {
     Assert.assertTrue("", fsView.getTrashRoots(true).size() > 0);
   }
 
-  // Default implementation of getTrashRoot for a fallback FS mounted at root:
-  // e.g., fallbackFS.uri.getPath = '/'
-  Path getTrashRootInFallBackFS() throws IOException {
-    return new Path(fsTarget.getHomeDirectory().toUri().getPath(),
-        TRASH_PREFIX);
-  }
+  @Test(expected = NotInMountpointException.class)
+  public void testViewFileSystemUtil() throws Exception {
+    Configuration newConf = new Configuration(conf);
 
-  /**
-   * Test TRASH_FORCE_INSIDE_MOUNT_POINT feature for getTrashRoot.
-   */
-  @Test
-  public void testTrashRootForceInsideMountPoint() throws IOException {
-    UserGroupInformation ugi = UserGroupInformation.getCurrentUser();
-    Configuration conf2 = new Configuration(conf);
-    conf2.setBoolean(CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT, true);
-    ConfigUtil.addLinkFallback(conf2, targetTestRoot.toUri());
-    FileSystem fsView2 = FileSystem.get(FsConstants.VIEWFS_URI, conf2);
+    FileSystem fileSystem = FileSystem.get(FsConstants.LOCAL_FS_URI,
+        newConf);
+    Assert.assertFalse("Unexpected FileSystem: " + fileSystem,
+        ViewFileSystemUtil.isViewFileSystem(fileSystem));
 
-    // Case 1: path p in the /data mount point.
-    // Return a trash root within the /data mount point.
-    Path dataTestPath = new Path("/data/dir/file");
-    Path dataTrashRoot = fsView2.makeQualified(
-        new Path("/data/" + TRASH_PREFIX + "/" + ugi.getShortUserName()));
-    Assert.assertEquals(dataTrashRoot, fsView2.getTrashRoot(dataTestPath));
+    fileSystem = FileSystem.get(FsConstants.VIEWFS_URI,
+        newConf);
+    Assert.assertTrue("Unexpected FileSystem: " + fileSystem,
+        ViewFileSystemUtil.isViewFileSystem(fileSystem));
 
-    // Case 2: path p not found in mount table.
-    // Return a trash root in fallback FS.
-    Path nonExistentPath = new Path("/nonExistentDir/nonExistentFile");
-    Path expectedTrash =
-        fsView2.makeQualified(getTrashRootInFallBackFS());
-    Assert.assertEquals(expectedTrash, fsView2.getTrashRoot(nonExistentPath));
+    // Case 1: Verify FsStatus of root path returns all MountPoints status.
+    Map<MountPoint, FsStatus> mountPointFsStatusMap =
+        ViewFileSystemUtil.getStatus(fileSystem, InodeTree.SlashPath);
+    Assert.assertEquals(getExpectedMountPoints(), mountPointFsStatusMap.size());
 
-    // Case 3: turn off the CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT flag.
-    // Return a trash root in user home dir.
-    conf2.setBoolean(CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT, false);
-    fsView2 = FileSystem.get(FsConstants.VIEWFS_URI, conf2);
-    Path targetFSUserHomeTrashRoot = fsTarget.makeQualified(
-        new Path(fsTarget.getHomeDirectory(), TRASH_PREFIX));
-    Assert.assertEquals(targetFSUserHomeTrashRoot,
-        fsView2.getTrashRoot(dataTestPath));
+    // Case 2: Verify FsStatus of an internal dir returns all
+    // MountPoints status.
+    mountPointFsStatusMap =
+        ViewFileSystemUtil.getStatus(fileSystem, new Path("/internalDir"));
+    Assert.assertEquals(getExpectedMountPoints(), mountPointFsStatusMap.size());
 
-    // Case 4: viewFS without fallback. Expect exception for a nonExistent path
-    conf2 = new Configuration(conf);
-    fsView2 = FileSystem.get(FsConstants.VIEWFS_URI, conf2);
-    try {
-      fsView2.getTrashRoot(nonExistentPath);
-    } catch (NotInMountpointException ignored) {
+    // Case 3: Verify FsStatus of a matching MountPoint returns exactly
+    // the corresponding MountPoint status.
+    mountPointFsStatusMap =
+        ViewFileSystemUtil.getStatus(fileSystem, new Path("/user"));
+    Assert.assertEquals(1, mountPointFsStatusMap.size());
+    for (Entry<MountPoint, FsStatus> entry : mountPointFsStatusMap.entrySet()) {
+      Assert.assertEquals(entry.getKey().getMountedOnPath().toString(),
+          "/user");
     }
-  }
 
-  /**
-   * A mocked FileSystem which returns a deep trash dir.
-   */
-  static class DeepTrashRootMockFS extends MockFileSystem {
-    public static final Path TRASH =
-        new Path("/vol/very/deep/deep/trash/dir/.Trash");
-
-    @Override
-    public Path getTrashRoot(Path path) {
-      return TRASH;
+    // Case 4: Verify FsStatus of a path over a MountPoint returns the
+    // corresponding MountPoint status.
+    mountPointFsStatusMap =
+        ViewFileSystemUtil.getStatus(fileSystem, new Path("/user/cloud"));
+    Assert.assertEquals(1, mountPointFsStatusMap.size());
+    for (Entry<MountPoint, FsStatus> entry : mountPointFsStatusMap.entrySet()) {
+      Assert.assertEquals(entry.getKey().getMountedOnPath().toString(),
+          "/user");
     }
-  }
 
-  /**
-   * Test getTrashRoot that is very deep inside a mount point.
-   */
-  @Test
-  public void testTrashRootDeepTrashDir() throws IOException {
+    // Case 5: Verify FsStatus of any level of an internal dir
+    // returns all MountPoints status.
+    mountPointFsStatusMap =
+        ViewFileSystemUtil.getStatus(fileSystem,
+            new Path("/internalDir/internalDir2"));
+    Assert.assertEquals(getExpectedMountPoints(), mountPointFsStatusMap.size());
 
-    Configuration conf2 = ViewFileSystemTestSetup.createConfig();
-    conf2.setBoolean(CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT, true);
-    conf2.setClass("fs.mocktrashfs.impl", DeepTrashRootMockFS.class,
-        FileSystem.class);
-    ConfigUtil.addLink(conf2, "/mnt/datavol1",
-        URI.create("mocktrashfs://localhost/vol"));
-    Path testPath = new Path("/mnt/datavol1/projs/proj");
-    FileSystem fsView2 = FileSystem.get(FsConstants.VIEWFS_URI, conf2);
-    Path expectedTrash = fsView2.makeQualified(
-        new Path("/mnt/datavol1/very/deep/deep/trash/dir/.Trash"));
-    Assert.assertEquals(expectedTrash, fsView2.getTrashRoot(testPath));
-  }
+    // Case 6: Verify FsStatus of a MountPoint URI returns
+    // the MountPoint status.
+    mountPointFsStatusMap =
+        ViewFileSystemUtil.getStatus(fileSystem, new Path("viewfs:/user/"));
+    Assert.assertEquals(1, mountPointFsStatusMap.size());
+    for (Entry<MountPoint, FsStatus> entry : mountPointFsStatusMap.entrySet()) {
+      Assert.assertEquals(entry.getKey().getMountedOnPath().toString(),
+          "/user");
+    }
 
-  /**
-   * Test getTrashRoots() for all users.
-   */
-  @Test
-  public void testTrashRootsAllUsers() throws IOException {
-    Configuration conf2 = new Configuration(conf);
-    conf2.setBoolean(CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT, true);
-    FileSystem fsView2 = FileSystem.get(FsConstants.VIEWFS_URI, conf2);
-
-    // Case 1: verify correct trash roots from fsView and fsView2
-    int beforeTrashRootNum = fsView.getTrashRoots(true).size();
-    int beforeTrashRootNum2 = fsView2.getTrashRoots(true).size();
-    Assert.assertEquals(beforeTrashRootNum, beforeTrashRootNum2);
-
-    fsView.mkdirs(new Path("/data/" + TRASH_PREFIX + "/user1"));
-    fsView.mkdirs(new Path("/data/" + TRASH_PREFIX + "/user2"));
-    fsView.mkdirs(new Path("/user/" + TRASH_PREFIX + "/user3"));
-    fsView.mkdirs(new Path("/user/" + TRASH_PREFIX + "/user4"));
-    fsView.mkdirs(new Path("/user2/" + TRASH_PREFIX + "/user5"));
-    int afterTrashRootsNum = fsView.getTrashRoots(true).size();
-    int afterTrashRootsNum2 = fsView2.getTrashRoots(true).size();
-    Assert.assertEquals(beforeTrashRootNum, afterTrashRootsNum);
-    Assert.assertEquals(beforeTrashRootNum2 + 5, afterTrashRootsNum2);
-
-    // Case 2: per-user mount point
-    fsTarget.mkdirs(new Path(targetTestRoot, "Users/userA/.Trash/userA"));
-    Configuration conf3 = new Configuration(conf2);
-    ConfigUtil.addLink(conf3, "/Users/userA",
-        new Path(targetTestRoot, "Users/userA").toUri());
-    FileSystem fsView3 = FileSystem.get(FsConstants.VIEWFS_URI, conf3);
-    int trashRootsNum3 = fsView3.getTrashRoots(true).size();
-    Assert.assertEquals(afterTrashRootsNum2 + 1, trashRootsNum3);
-
-    // Case 3: single /Users mount point for all users
-    fsTarget.mkdirs(new Path(targetTestRoot, "Users/.Trash/user1"));
-    fsTarget.mkdirs(new Path(targetTestRoot, "Users/.Trash/user2"));
-    Configuration conf4 = new Configuration(conf2);
-    ConfigUtil.addLink(conf4, "/Users",
-        new Path(targetTestRoot, "Users").toUri());
-    FileSystem fsView4 = FileSystem.get(FsConstants.VIEWFS_URI, conf4);
-    int trashRootsNum4 = fsView4.getTrashRoots(true).size();
-    Assert.assertEquals(afterTrashRootsNum2 + 2, trashRootsNum4);
-
-    // Case 4: test trash roots in fallback FS
-    fsTarget.mkdirs(new Path(targetTestRoot, ".Trash/user10"));
-    fsTarget.mkdirs(new Path(targetTestRoot, ".Trash/user11"));
-    fsTarget.mkdirs(new Path(targetTestRoot, ".Trash/user12"));
-    Configuration conf5 = new Configuration(conf2);
-    ConfigUtil.addLinkFallback(conf5, targetTestRoot.toUri());
-    FileSystem fsView5 = FileSystem.get(FsConstants.VIEWFS_URI, conf5);
-    int trashRootsNum5 = fsView5.getTrashRoots(true).size();
-    Assert.assertEquals(afterTrashRootsNum2 + 3, trashRootsNum5);
-  }
-
-  /**
-   * Test getTrashRoots() for current user.
-   */
-  @Test
-  public void testTrashRootsCurrentUser() throws IOException {
-    String currentUser =
-        UserGroupInformation.getCurrentUser().getShortUserName();
-    Configuration conf2 = new Configuration(conf);
-    conf2.setBoolean(CONFIG_VIEWFS_TRASH_FORCE_INSIDE_MOUNT_POINT, true);
-    FileSystem fsView2 = FileSystem.get(FsConstants.VIEWFS_URI, conf2);
-
-    int beforeTrashRootNum = fsView.getTrashRoots(false).size();
-    int beforeTrashRootNum2 = fsView2.getTrashRoots(false).size();
-    Assert.assertEquals(beforeTrashRootNum, beforeTrashRootNum2);
-
-    fsView.mkdirs(new Path("/data/" + TRASH_PREFIX + "/" + currentUser));
-    fsView.mkdirs(new Path("/data/" + TRASH_PREFIX + "/user2"));
-    fsView.mkdirs(new Path("/user/" + TRASH_PREFIX + "/" + currentUser));
-    fsView.mkdirs(new Path("/user/" + TRASH_PREFIX + "/user4"));
-    fsView.mkdirs(new Path("/user2/" + TRASH_PREFIX + "/user5"));
-    int afterTrashRootsNum = fsView.getTrashRoots(false).size();
-    int afterTrashRootsNum2 = fsView2.getTrashRoots(false).size();
-    Assert.assertEquals(beforeTrashRootNum, afterTrashRootsNum);
-    Assert.assertEquals(beforeTrashRootNum2 + 2, afterTrashRootsNum2);
-
-    // Test trash roots in fallback FS
-    Configuration conf3 = new Configuration(conf2);
-    fsTarget.mkdirs(new Path(targetTestRoot, TRASH_PREFIX + "/" + currentUser));
-    ConfigUtil.addLinkFallback(conf3, targetTestRoot.toUri());
-    FileSystem fsView3 = FileSystem.get(FsConstants.VIEWFS_URI, conf3);
-    int trashRootsNum3 = fsView3.getTrashRoots(false).size();
-    Assert.assertEquals(afterTrashRootsNum2 + 1, trashRootsNum3);
+    // Case 7: Verify FsStatus of a non MountPoint path throws exception
+    ViewFileSystemUtil.getStatus(fileSystem, new Path("/non-existing"));
   }
 
   @Test
@@ -1211,9 +1186,111 @@ abstract public class ViewFileSystemBaseTest {
     });
   }
 
+  @Test
+  public void testUsed() throws IOException {
+    try {
+      fsView.getUsed();
+      fail("ViewFileSystem getUsed() should fail for slash root path when the" +
+          " slash root mount point is not configured.");
+    } catch (NotInMountpointException e) {
+      // expected exception.
+    }
+    long usedSpaceByPathViaViewFs = fsView.getUsed(new Path("/user"));
+    long usedSpaceByPathViaTargetFs =
+        fsTarget.getUsed(new Path(targetTestRoot, "user"));
+    assertEquals("Space used not matching between ViewFileSystem and " +
+        "the mounted FileSystem!",
+        usedSpaceByPathViaTargetFs, usedSpaceByPathViaViewFs);
+
+    Path mountDataRootPath = new Path("/data");
+    String fsTargetFileName = "debug.log";
+    Path fsTargetFilePath = new Path(targetTestRoot, "data/debug.log");
+    Path mountDataFilePath = new Path(mountDataRootPath, fsTargetFileName);
+    fileSystemTestHelper.createFile(fsTarget, fsTargetFilePath);
+
+    usedSpaceByPathViaViewFs = fsView.getUsed(mountDataFilePath);
+    usedSpaceByPathViaTargetFs = fsTarget.getUsed(fsTargetFilePath);
+    assertEquals("Space used not matching between ViewFileSystem and " +
+        "the mounted FileSystem!",
+        usedSpaceByPathViaTargetFs, usedSpaceByPathViaViewFs);
+  }
+
+  @Test
+  public void testLinkTarget() throws Exception {
+
+    Assume.assumeTrue(fsTarget.supportsSymlinks() &&
+        fsTarget.areSymlinksEnabled());
+
+    // Symbolic link
+    final String targetFileName = "debug.log";
+    final String linkFileName = "debug.link";
+    final Path targetFile = new Path(targetTestRoot, targetFileName);
+    final Path symLink = new Path(targetTestRoot, linkFileName);
+
+    FileSystemTestHelper.createFile(fsTarget, targetFile);
+    fsTarget.createSymlink(targetFile, symLink, false);
+
+    final Path mountTargetRootPath = new Path("/targetRoot");
+    final Path mountTargetSymLinkPath = new Path(mountTargetRootPath,
+        linkFileName);
+    final Path expectedMountLinkTarget = fsTarget.makeQualified(
+        new Path(targetTestRoot, targetFileName));
+    final Path actualMountLinkTarget = fsView.getLinkTarget(
+        mountTargetSymLinkPath);
+
+    assertEquals("Resolved link target path not matching!",
+        expectedMountLinkTarget, actualMountLinkTarget);
+
+    // Relative symbolic link
+    final String relativeFileName = "dir2/../" + targetFileName;
+    final String link2FileName = "dir2/rel.link";
+    final Path relTargetFile = new Path(targetTestRoot, relativeFileName);
+    final Path relativeSymLink = new Path(targetTestRoot, link2FileName);
+    fsTarget.createSymlink(relTargetFile, relativeSymLink, true);
+
+    final Path mountTargetRelativeSymLinkPath = new Path(mountTargetRootPath,
+        link2FileName);
+    final Path expectedMountRelLinkTarget = fsTarget.makeQualified(
+        new Path(targetTestRoot, relativeFileName));
+    final Path actualMountRelLinkTarget = fsView.getLinkTarget(
+        mountTargetRelativeSymLinkPath);
+
+    assertEquals("Resolved relative link target path not matching!",
+        expectedMountRelLinkTarget, actualMountRelLinkTarget);
+
+    try {
+      fsView.getLinkTarget(new Path("/linkToAFile"));
+      fail("Resolving link target for a ViewFs mount link should fail!");
+    } catch (Exception e) {
+      LOG.info("Expected exception: " + e);
+      assertThat(e.getMessage(),
+          containsString("not a symbolic link"));
+    }
+
+    try {
+      fsView.getLinkTarget(fsView.makeQualified(
+          new Path(mountTargetRootPath, targetFileName)));
+      fail("Resolving link target for a non sym link should fail!");
+    } catch (Exception e) {
+      LOG.info("Expected exception: " + e);
+      assertThat(e.getMessage(),
+          containsString("not a symbolic link"));
+    }
+
+    try {
+      fsView.getLinkTarget(new Path("/targetRoot/non-existing-file"));
+      fail("Resolving link target for a non existing link should fail!");
+    } catch (Exception e) {
+      LOG.info("Expected exception: " + e);
+      assertThat(e.getMessage(),
+          containsString("File does not exist:"));
+    }
+  }
+
   @Test
   public void testViewFileSystemInnerCache() throws Exception {
-    ViewFileSystem.InnerCache cache = new ViewFileSystem.InnerCache();
+    ViewFileSystem.InnerCache cache =
+        new ViewFileSystem.InnerCache(new FsGetter());
     FileSystem fs = cache.get(fsTarget.getUri(), conf);
 
     // InnerCache caches filesystem.
@@ -1305,28 +1382,55 @@ abstract public class ViewFileSystemBaseTest {
     assertFalse(fsTarget.exists(realTestPath));
   }
 
-  public void testConfLinkSlash() throws Exception {
-    String clusterName = "ClusterX";
-    URI viewFsUri = new URI(FsConstants.VIEWFS_SCHEME, clusterName,
-        "/", null, null);
+  @Test
+  public void testGetContentSummary() throws IOException {
+    ContentSummary summaryBefore =
+        fsView.getContentSummary(new Path("/internalDir"));
+    String expected = "GET CONTENT SUMMARY";
+    Path filePath =
+        new Path("/internalDir/internalDir2/linkToDir3", "foo");
 
-    Configuration newConf = new Configuration();
-    ConfigUtil.addLink(newConf, clusterName, "/",
-        new Path(targetTestRoot, "/").toUri());
+    try (FSDataOutputStream outputStream = fsView.create(filePath)) {
+      outputStream.write(expected.getBytes());
+    }
 
-    String mtPrefix = Constants.CONFIG_VIEWFS_PREFIX + "." + clusterName + ".";
-    try {
-      FileSystem.get(viewFsUri, newConf);
-      fail("ViewFileSystem should error out on mount table entry: "
-          + mtPrefix + Constants.CONFIG_VIEWFS_LINK + "." + "/");
-    } catch (Exception e) {
-      if (e instanceof UnsupportedFileSystemException) {
-        String msg = " Use " + Constants.CONFIG_VIEWFS_LINK_MERGE_SLASH +
-            " instead";
-        assertThat(e.getMessage(), containsString(msg));
-      } else {
-        fail("Unexpected exception: " + e.getMessage());
-      }
+    Path newDirPath = new Path("/internalDir/linkToDir2", "bar");
+    fsView.mkdirs(newDirPath);
+
+    ContentSummary summaryAfter =
+        fsView.getContentSummary(new Path("/internalDir"));
+    assertEquals("The file count didn't match",
+        summaryBefore.getFileCount() + 1,
+        summaryAfter.getFileCount());
+    assertEquals("The size didn't match",
+        summaryBefore.getLength() + expected.length(),
+        summaryAfter.getLength());
+    assertEquals("The directory count didn't match",
+        summaryBefore.getDirectoryCount() + 1,
+        summaryAfter.getDirectoryCount());
+  }
+
+  @Test
+  public void testGetContentSummaryWithFileInLocalFS() throws Exception {
+    ContentSummary summaryBefore =
+        fsView.getContentSummary(new Path("/internalDir"));
+    String expected = "GET CONTENT SUMMARY";
+    File localFile = temporaryFolder.newFile("localFile");
+    try (FileOutputStream fos = new FileOutputStream(localFile)) {
+      fos.write(expected.getBytes());
+    }
+    ConfigUtil.addLink(conf,
+        "/internalDir/internalDir2/linkToLocalFile", localFile.toURI());
+
+    try (FileSystem fs = FileSystem.get(FsConstants.VIEWFS_URI, conf)) {
+      ContentSummary summaryAfter =
+          fs.getContentSummary(new Path("/internalDir"));
+      assertEquals("The file count didn't match",
+          summaryBefore.getFileCount() + 1,
+          summaryAfter.getFileCount());
+      assertEquals("The directory count didn't match",
+          summaryBefore.getLength() + expected.length(),
+          summaryAfter.getLength());
     }
   }
 
@@ -1370,65 +1474,4 @@ abstract public class ViewFileSystemBaseTest {
     // viewfs inner cache is disabled
     assertEquals(cacheSize + 2, TestFileUtil.getCacheSize());
   }
-
-  @Test
-  public void testTargetFileSystemLazyInitializationForChecksumMethods()
-      throws Exception {
-    final String clusterName = "cluster" + new Random().nextInt();
-    Configuration config = new Configuration(conf);
-    config.setBoolean(CONFIG_VIEWFS_ENABLE_INNER_CACHE, false);
-    config.setClass("fs.othermockfs.impl",
-        TestChRootedFileSystem.MockFileSystem.class, FileSystem.class);
-    ConfigUtil.addLink(config, clusterName, "/user",
-        URI.create("othermockfs://mockauth1/mockpath"));
-    ConfigUtil.addLink(config, clusterName,
-        "/mock", URI.create("othermockfs://mockauth/mockpath"));
-
-    final int cacheSize = TestFileUtil.getCacheSize();
-    ViewFileSystem viewFs = (ViewFileSystem) FileSystem.get(
-        new URI("viewfs://" + clusterName + "/"), config);
-
-    // As no inner file system instance has been initialized,
-    // cache size will remain the same
-    // cache is disabled for viewfs scheme, so the viewfs:// instance won't
-    // go in the cache even after the initialization
-    assertEquals(cacheSize, TestFileUtil.getCacheSize());
-
-    // This is not going to initialize any filesystem instance
-    viewFs.setVerifyChecksum(true);
-
-    // Cache size will remain the same
-    assertEquals(cacheSize, TestFileUtil.getCacheSize());
-
-    // This resolve path will initialize the file system corresponding
-    // to the mount table entry of the path "/user"
-    viewFs.getFileChecksum(
-        new Path(String.format("viewfs://%s/%s", clusterName, "/user")));
-
-    // Cache size will increase by 1.
-    assertEquals(cacheSize + 1, TestFileUtil.getCacheSize());
-
-    viewFs.close();
-    // Initialized FileSystem instances will not be removed from cache as
-    // viewfs inner cache is disabled
-    assertEquals(cacheSize + 1, TestFileUtil.getCacheSize());
-  }
-
-  @Test
-  public void testInvalidMountPoints() throws Exception {
-    final String clusterName = "cluster" + new Random().nextInt();
-    Configuration config = new Configuration(conf);
-    config.set(ConfigUtil.getConfigViewFsPrefix(clusterName) + "." +
-        Constants.CONFIG_VIEWFS_LINK + "." + "/invalidPath",
-        "othermockfs:|mockauth/mockpath");
-
-    try {
-      FileSystem viewFs = FileSystem.get(
-          new URI("viewfs://" + clusterName + "/"), config);
-      fail("FileSystem should not initialize. Should fail with IOException");
-    } catch (IOException ex) {
-      assertTrue("Should get URISyntax Exception",
-          ex.getMessage().startsWith("URISyntax exception"));
-    }
-  }
 }
