@@ -21,12 +21,14 @@ import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
+import java.net.URI;
 import java.nio.channels.ClosedChannelException;
 import java.util.ArrayList;
+import java.util.Collection;
 import java.util.Collections;
 import java.util.HashMap;
-import java.util.HashSet;
 import java.util.LinkedList;
+import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
@@ -38,22 +40,25 @@ import javax.management.ObjectName;
 import javax.management.StandardMBean;
 
 import com.google.common.math.LongMath;
-import org.apache.commons.lang.ArrayUtils;
+import org.apache.commons.lang3.ArrayUtils;
 import org.apache.hadoop.conf.Configuration;
+import org.apache.hadoop.fs.DF;
 import org.apache.hadoop.fs.StorageType;
 import org.apache.hadoop.hdfs.server.datanode.checker.VolumeCheckResult;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.impl.FsDatasetImplTestUtils;
+import org.apache.hadoop.util.AutoCloseableLock;
 import org.apache.hadoop.hdfs.DFSConfigKeys;
 import org.apache.hadoop.hdfs.protocol.Block;
 import org.apache.hadoop.hdfs.protocol.BlockListAsLongs;
 import org.apache.hadoop.hdfs.protocol.BlockLocalPathInfo;
 import org.apache.hadoop.hdfs.protocol.ExtendedBlock;
-import org.apache.hadoop.hdfs.protocol.HdfsBlocksMetadata;
 import org.apache.hadoop.hdfs.server.common.HdfsServerConstants.ReplicaState;
+import org.apache.hadoop.hdfs.server.datanode.DirectoryScanner.ReportCompiler;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.DataNodeVolumeMetrics;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.FsDatasetSpi;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.FsVolumeReference;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.FsVolumeSpi;
+import org.apache.hadoop.hdfs.server.datanode.fsdataset.FsVolumeSpi.ScanInfo;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.LengthInputStream;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.ReplicaInputStreams;
 import org.apache.hadoop.hdfs.server.datanode.fsdataset.ReplicaOutputStreams;
@@ -68,23 +73,22 @@ import org.apache.hadoop.hdfs.server.protocol.VolumeFailureSummary;
 import org.apache.hadoop.io.IOUtils;
 import org.apache.hadoop.metrics2.MetricsCollector;
 import org.apache.hadoop.metrics2.util.MBeans;
-import org.apache.hadoop.util.AutoCloseableLock;
 import org.apache.hadoop.util.DataChecksum;
 
 /**
  * This class implements a simulated FSDataset.
- *
+ * 
  * Blocks that are created are recorded but their data (plus their CRCs) are
  *  discarded.
  * Fixed data is returned when blocks are read; a null CRC meta file is
  * created for such data.
- *
+ * 
  * This FSDataset does not remember any block information across its
  * restarts; it does however offer an operation to inject blocks
  *  (See the TestInectionForSImulatedStorage()
  * for a usage example of injection.
- *
- * Note the synchronization is coarse grained - it is at each method.
+ * 
+ * Note the synchronization is coarse grained - it is at each method. 
  */
 public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   public final static int BYTE_MASK = 0xff;
@@ -141,23 +145,24 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
 
   public static byte simulatedByte(Block b, long offsetInBlk) {
     byte firstByte = (byte) (b.getBlockId() & BYTE_MASK);
-    return (byte) ((firstByte + offsetInBlk) & BYTE_MASK);
+    return (byte) ((firstByte + offsetInBlk % 29) & BYTE_MASK);
   }
-
+  
   public static final String CONFIG_PROPERTY_CAPACITY =
       "dfs.datanode.simulateddatastorage.capacity";
-
+  
   public static final long DEFAULT_CAPACITY = 2L<<40; // 1 terabyte
-
+  
   public static final String CONFIG_PROPERTY_STATE =
       "dfs.datanode.simulateddatastorage.state";
   private static final DatanodeStorage.State DEFAULT_STATE =
       DatanodeStorage.State.NORMAL;
+  
+  static final byte[] nullCrcFileData;
 
   private final AutoCloseableLock datasetLock;
   private final FileIoProvider fileIoProvider;
 
-  static final byte[] nullCrcFileData;
   static {
     DataChecksum checksum = DataChecksum.newDataChecksum(
         DataChecksum.Type.NULL, 16*1024 );
@@ -171,7 +176,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   }
 
   // information about a single block
-  private class BInfo implements ReplicaInPipelineInterface {
+  private class BInfo implements ReplicaInPipeline {
     final Block theBlock;
     private boolean finalized = false; // if not finalized => ongoing creation
     SimulatedOutputStream oStream = null;
@@ -198,7 +203,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
         oStream = null;
       }
     }
-
+    
     @Override
     public String getStorageUuid() {
       return getStorage(theBlock).getStorageUuid();
@@ -226,7 +231,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
         theBlock.setNumBytes(length);
       }
     }
-
+    
     synchronized SimulatedInputStream getIStream() {
       if (!finalized) {
         // throw new IOException("Trying to read an unfinalized block");
@@ -235,12 +240,12 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
         return new SimulatedInputStream(theBlock.getNumBytes(), theBlock);
       }
     }
-
+    
     synchronized void finalizeBlock(String bpid, long finalSize)
         throws IOException {
       if (finalized) {
         throw new IOException(
-            "Finalizing a block that has already been finalized" +
+            "Finalizing a block that has already been finalized" + 
             theBlock.getBlockId());
       }
       if (oStream == null) {
@@ -254,7 +259,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
         throw new IOException(
           "Size passed to finalize does not match the amount of data written");
       }
-      // We had allocated the expected length when block was created;
+      // We had allocated the expected length when block was created; 
       // adjust if necessary
       long extraLen = finalSize - theBlock.getNumBytes();
       if (extraLen > 0) {
@@ -265,7 +270,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       } else {
         getStorage(theBlock).free(bpid, -extraLen);
       }
-      theBlock.setNumBytes(finalSize);
+      theBlock.setNumBytes(finalSize);  
 
       finalized = true;
       oStream = null;
@@ -286,7 +291,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     }
 
     SimulatedInputStream getMetaIStream() {
-      return new SimulatedInputStream(nullCrcFileData);
+      return new SimulatedInputStream(nullCrcFileData);  
     }
 
     synchronized boolean isFinalized() {
@@ -294,8 +299,9 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     }
 
     @Override
-    synchronized public ReplicaOutputStreams createStreams(boolean isCreate,
-        DataChecksum requestedChecksum) throws IOException {
+    synchronized public ReplicaOutputStreams createStreams(boolean isCreate, 
+        DataChecksum requestedChecksum)
+        throws IOException {
       if (finalized) {
         throw new IOException("Trying to write to a finalized replica "
             + theBlock);
@@ -369,8 +375,30 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     public boolean isOnTransientStorage() {
       return false;
     }
-  }
 
+    @Override
+    public ReplicaInfo getReplicaInfo() {
+      return null;
+    }
+
+    @Override
+    public void setWriter(Thread writer) {
+    }
+
+    @Override
+    public void interruptThread() {
+    }
+
+    @Override
+    public boolean attemptToSetWriter(Thread prevWriter, Thread newWriter) {
+      return false;
+    }
+
+    @Override
+    public void stopWriter(long xceiverStopTimeout) throws IOException {
+    }
+  }
+  
   /**
    * Class is used for tracking block pool storage utilization similar
    * to {@link BlockPoolSlice}
@@ -378,48 +406,48 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   private static class SimulatedBPStorage {
     private long used;    // in bytes
     private final Map<Block, BInfo> blockMap = new TreeMap<>();
-
+    
     long getUsed() {
       return used;
     }
-
+    
     void alloc(long amount) {
       used += amount;
     }
-
+    
     void free(long amount) {
       used -= amount;
     }
-
+    
     Map<Block, BInfo> getBlockMap() {
       return blockMap;
     }
 
     SimulatedBPStorage() {
-      used = 0;
+      used = 0;   
     }
   }
-
+  
   /**
    * Class used for tracking datanode level storage utilization similar
    * to {@link FSVolumeSet}
    */
-  private static class SimulatedStorage {
+  static class SimulatedStorage {
     private final Map<String, SimulatedBPStorage> map =
         new ConcurrentHashMap<>();
 
     private final long capacity;  // in bytes
     private final DatanodeStorage dnStorage;
     private final SimulatedVolume volume;
-
+    
     synchronized long getFree() {
       return capacity - getUsed();
     }
-
+    
     long getCapacity() {
       return capacity;
     }
-
+    
     synchronized long getUsed() {
       long used = 0;
       for (SimulatedBPStorage bpStorage : map.values()) {
@@ -427,11 +455,11 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       }
       return used;
     }
-
+    
     synchronized long getBlockPoolUsed(String bpid) throws IOException {
       return getBPStorage(bpid).getUsed();
     }
-
+    
     int getNumFailedVolumes() {
       return 0;
     }
@@ -441,13 +469,13 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
         getBPStorage(bpid).alloc(amount);
         return true;
       }
-      return false;
+      return false;    
     }
-
+    
     synchronized void free(String bpid, long amount) throws IOException {
       getBPStorage(bpid).free(amount);
     }
-
+    
     SimulatedStorage(long cap, DatanodeStorage.State state,
         FileIoProvider fileIoProvider, Configuration conf) {
       capacity = cap;
@@ -458,7 +486,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
           DataNodeVolumeMetrics.create(conf, dnStorage.getStorageID());
       this.volume = new SimulatedVolume(this, fileIoProvider, volumeMetrics);
     }
-
+    
     synchronized void addBlockPool(String bpid) {
       SimulatedBPStorage bpStorage = map.get(bpid);
       if (bpStorage != null) {
@@ -466,11 +494,11 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       }
       map.put(bpid, new SimulatedBPStorage());
     }
-
+    
     synchronized void removeBlockPool(String bpid) {
       map.remove(bpid);
     }
-
+    
     private SimulatedBPStorage getBPStorage(String bpid) throws IOException {
       SimulatedBPStorage bpStorage = map.get(bpid);
       if (bpStorage == null) {
@@ -482,7 +510,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     String getStorageUuid() {
       return dnStorage.getStorageID();
     }
-
+    
     DatanodeStorage getDnStorage() {
       return dnStorage;
     }
@@ -505,7 +533,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       return bpStorage.getBlockMap();
     }
   }
-
+  
   static class SimulatedVolume implements FsVolumeSpi {
     private final SimulatedStorage storage;
     private final FileIoProvider fileIoProvider;
@@ -549,24 +577,9 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       return storage.getCapacity() - storage.getUsed();
     }
 
-    @Override
-    public String getBasePath() {
-      return null;
-    }
-
-    @Override
-    public String getPath(String bpid) throws IOException {
-      return null;
-    }
-
-    @Override
-    public File getFinalizedDir(String bpid) throws IOException {
-      return null;
-    }
-
     @Override
     public StorageType getStorageType() {
-      return null;
+      return StorageType.DISK;
     }
 
     @Override
@@ -602,12 +615,38 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       throw new UnsupportedOperationException();
     }
 
+    @Override
+    public StorageLocation getStorageLocation() {
+      try {
+        return StorageLocation.parse("[DISK]file:///simulated");
+      } catch (Exception e) {
+        return null;
+      }
+    }
+
+    @Override
+    public URI getBaseURI() {
+      return null;
+    }
+
+    @Override
+    public DF getUsageStats(Configuration conf) {
+      return null;
+    }
+
     @Override
     public byte[] loadLastPartialChunkChecksum(
         File blockFile, File metaFile) throws IOException {
       return null;
     }
 
+    @Override
+    public LinkedList<ScanInfo> compileReport(String bpid,
+        LinkedList<ScanInfo> report, ReportCompiler reportCompiler)
+        throws InterruptedException, IOException {
+      return null;
+    }
+
     @Override
     public FileIoProvider getFileIoProvider() {
       return fileIoProvider;
@@ -628,7 +667,11 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   private final List<SimulatedStorage> storages;
   private final String datanodeUuid;
   private final DataNode datanode;
+  
 
+  public List<SimulatedStorage> getStorages() {
+    return storages;
+  }
 
   public SimulatedFSDataset(DataStorage storage, Configuration conf) {
     this(null, storage, conf);
@@ -640,7 +683,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     if (storage != null && storage.getNumStorageDirs() > 0) {
       storageCount = storage.getNumStorageDirs();
       for (int i = 0; i < storage.getNumStorageDirs(); ++i) {
-        DataStorage.createStorageID(storage.getStorageDir(i), false);
+        DataStorage.createStorageID(storage.getStorageDir(i), false, conf);
       }
       this.datanodeUuid = storage.getDatanodeUuid();
     } else {
@@ -788,7 +831,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     }
     return total;
   }
-
+  
   @Override // FSDatasetMBean
   public long getRemaining() {
 
@@ -889,7 +932,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     }
   }
 
-  @Override
+  @Override 
   public synchronized String getReplicaString(String bpid, long blockId) {
     Replica r = null;
     try {
@@ -985,14 +1028,14 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
    *
    * @throws ReplicaNotFoundException          If the replica is not found
    *
-   * @throws UnexpectedReplicaStateException   If the replica is not in the
+   * @throws UnexpectedReplicaStateException   If the replica is not in the 
    *                                             expected state.
    */
   @Override // {@link FsDatasetSpi}
   public void checkBlock(ExtendedBlock b, long minLength, ReplicaState state)
       throws ReplicaNotFoundException, UnexpectedReplicaStateException {
     final BInfo binfo = getBInfo(b);
-
+    
     if (binfo == null) {
       throw new ReplicaNotFoundException(b);
     }
@@ -1075,7 +1118,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     map.put(binfo.theBlock, binfo);
     return binfo;
   }
-
+  
   @Override // FsDatasetSpi
   public synchronized ReplicaHandler recoverRbw(
       ExtendedBlock b, long newGS, long minBytesRcvd, long maxBytesRcvd)
@@ -1098,37 +1141,38 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
 
   @Override // FsDatasetSpi
   public synchronized ReplicaHandler createRbw(
-      StorageType storageType, ExtendedBlock b,
+      StorageType storageType, String storageId, ExtendedBlock b,
       boolean allowLazyPersist) throws IOException {
-    return createTemporary(storageType, b, false);
+    return createTemporary(storageType, storageId, b, false);
   }
 
   @Override // FsDatasetSpi
   public synchronized ReplicaHandler createTemporary(StorageType storageType,
-      ExtendedBlock b, boolean isTransfer) throws IOException {
+      String storageId, ExtendedBlock b, boolean isTransfer)
+      throws IOException {
     if (isValidBlock(b)) {
-          throw new ReplicaAlreadyExistsException("Block " + b +
-              " is valid, and cannot be written to.");
-      }
+      throw new ReplicaAlreadyExistsException("Block " + b +
+          " is valid, and cannot be written to.");
+    }
     if (isValidRbw(b)) {
-        throw new ReplicaAlreadyExistsException("Block " + b +
-            " is being written, and cannot be written to.");
+      throw new ReplicaAlreadyExistsException("Block " + b +
+          " is being written, and cannot be written to.");
     }
     BInfo binfo = new BInfo(b.getBlockPoolId(), b.getLocalBlock(), true);
     getBlockMap(b).put(binfo.theBlock, binfo);
     return new ReplicaHandler(binfo, null);
   }
 
-  protected synchronized InputStream getBlockInputStream(ExtendedBlock b)
+  public synchronized InputStream getBlockInputStream(ExtendedBlock b)
       throws IOException {
     BInfo binfo = getBlockMap(b).get(b.getLocalBlock());
     if (binfo == null) {
-      throw new IOException("No such Block " + b );
+      throw new IOException("No such Block " + b );  
     }
-
+    
     return binfo.getIStream();
   }
-
+  
   @Override // FsDatasetSpi
   public synchronized InputStream getBlockInputStream(ExtendedBlock b,
       long seekOffset) throws IOException {
@@ -1149,10 +1193,10 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       ) throws IOException {
     BInfo binfo = getBlockMap(b).get(b.getLocalBlock());
     if (binfo == null) {
-      throw new IOException("No such Block " + b );
+      throw new IOException("No such Block " + b );  
     }
     if (!binfo.finalized) {
-      throw new IOException("Block " + b +
+      throw new IOException("Block " + b + 
           " is being written, its meta cannot be read");
     }
     final SimulatedInputStream sin = binfo.getMetaIStream();
@@ -1165,12 +1209,12 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
 
   @Override // FsDatasetSpi
   public synchronized void adjustCrcChannelPosition(ExtendedBlock b,
-                                              ReplicaOutputStreams stream,
+                                              ReplicaOutputStreams stream, 
                                               int checksumSize)
                                               throws IOException {
   }
 
-  /**
+  /** 
    * Simulated input and output streams
    *
    */
@@ -1179,7 +1223,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     int currentPos = 0;
     byte[] data = null;
     Block theBlock = null;
-
+    
     /**
      * An input stream of size l with repeated bytes
      * @param l size of the stream
@@ -1189,7 +1233,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       length = l;
       theBlock = b;
     }
-
+    
     /**
      * An input stream of of the supplied data
      * @param iData data to construct the stream
@@ -1198,7 +1242,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       data = iData;
       length = data.length;
     }
-
+    
     /**
      * @return the lenght of the input stream
      */
@@ -1217,9 +1261,9 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
         return simulatedByte(theBlock, currentPos++) & BYTE_MASK;
       }
     }
-
+    
     @Override
-    public int read(byte[] b) throws IOException {
+    public int read(byte[] b) throws IOException { 
 
       if (b == null) {
         throw new NullPointerException();
@@ -1242,7 +1286,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       return bytesRead;
     }
   }
-
+  
   /**
    * This class implements an output stream that merely throws its data away, but records its
    * length.
@@ -1250,15 +1294,15 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
    */
   static private class SimulatedOutputStream extends OutputStream {
     long length = 0;
-
+    
     /**
      * constructor for Simulated Output Steram
      */
     SimulatedOutputStream() {
     }
-
+    
     /**
-     *
+     * 
      * @return the length of the data created so far.
      */
     long getLength() {
@@ -1270,17 +1314,17 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     void setLength(long length) {
       this.length = length;
     }
-
+    
     @Override
     public void write(int arg0) throws IOException {
       length++;
     }
-
+    
     @Override
     public void write(byte[] b) throws IOException {
       length += b.length;
     }
-
+    
     @Override
     public void write(byte[] b,
               int off,
@@ -1288,11 +1332,11 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       length += len;
     }
   }
-
+  
   private ObjectName mbeanName;
 
 
-
+  
   /**
    * Register the FSDataset MBean using the name
    *        "hadoop:service=DataNode,name=FSDatasetState-<storageid>"
@@ -1301,7 +1345,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
    */
   void registerMBean(final String storageId) {
     // We wrap to bypass standard mbean naming convetion.
-    // This wraping can be removed in java 6 as it is more flexible in
+    // This wraping can be removed in java 6 as it is more flexible in 
     // package naming for mbeans and their impl.
     StandardMBean bean;
 
@@ -1312,23 +1356,20 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     } catch (NotCompliantMBeanException e) {
       DataNode.LOG.warn("Error registering FSDatasetState MBean", e);
     }
-
+ 
     DataNode.LOG.info("Registered FSDatasetState MBean");
   }
 
   @Override
   public void shutdown() {
-    if (mbeanName != null) {
-      MBeans.unregister(mbeanName);
-      mbeanName = null;
-    }
+    if (mbeanName != null) MBeans.unregister(mbeanName);
   }
 
   @Override
   public String getStorageInfo() {
     return "Simulated FSDataset-" + datanodeUuid;
   }
-
+  
   @Override
   public boolean hasEnoughResource() {
     return true;
@@ -1340,11 +1381,11 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     ExtendedBlock b = rBlock.getBlock();
     BInfo binfo = getBlockMap(b).get(b.getLocalBlock());
     if (binfo == null) {
-      throw new IOException("No such Block " + b );
+      throw new IOException("No such Block " + b );  
     }
 
-    return new ReplicaRecoveryInfo(binfo.getBlockId(), binfo.getBytesOnDisk(),
-        binfo.getGenerationStamp(),
+    return new ReplicaRecoveryInfo(binfo.getBlockId(), binfo.getBytesOnDisk(), 
+        binfo.getGenerationStamp(), 
         binfo.isFinalized()?ReplicaState.FINALIZED : ReplicaState.RBW);
   }
 
@@ -1367,21 +1408,21 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
       storage.addBlockPool(bpid);
     }
   }
-
+  
   @Override // FsDatasetSpi
   public void shutdownBlockPool(String bpid) {
     for (SimulatedStorage storage : storages) {
       storage.removeBlockPool(bpid);
     }
   }
-
+  
   @Override // FsDatasetSpi
   public void deleteBlockPool(String bpid, boolean force) {
      return;
   }
 
   @Override
-  public ReplicaInPipelineInterface convertTemporaryToRbw(ExtendedBlock temporary)
+  public ReplicaInPipeline convertTemporaryToRbw(ExtendedBlock temporary)
       throws IOException {
     final BInfo r = getBlockMap(temporary).get(temporary.getLocalBlock());
     if (r == null) {
@@ -1398,12 +1439,6 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     throw new UnsupportedOperationException();
   }
 
-  @Override
-  public HdfsBlocksMetadata getHdfsBlocksMetadata(String bpid, long[] blockIds)
-      throws IOException {
-    throw new UnsupportedOperationException();
-  }
-
   @Override
   public void enableTrash(String bpid) {
     throw new UnsupportedOperationException();
@@ -1427,8 +1462,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   }
 
   @Override
-  public void checkAndUpdate(String bpid, long blockId, File diskFile,
-      File diskMetaFile, FsVolumeSpi vol) throws IOException {
+  public void checkAndUpdate(String bpid, ScanInfo info) throws IOException {
     throw new UnsupportedOperationException();
   }
 
@@ -1468,7 +1502,7 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   }
 
   @Override
-  public List<FinalizedReplica> getFinalizedBlocks(String bpid) {
+  public List<ReplicaInfo> getFinalizedBlocks(String bpid) {
     throw new UnsupportedOperationException();
   }
 
@@ -1483,7 +1517,8 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
   }
 
   @Override
-  public synchronized void removeVolumes(Set<File> volumes, boolean clearFailure) {
+  public synchronized void removeVolumes(Collection<StorageLocation> volumes,
+      boolean clearFailure) {
     throw new UnsupportedOperationException();
   }
 
@@ -1506,26 +1541,32 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
 
   @Override
   public ReplicaInfo moveBlockAcrossStorage(ExtendedBlock block,
-      StorageType targetStorageType) throws IOException {
+      StorageType targetStorageType, String storageId) throws IOException {
     // TODO Auto-generated method stub
     return null;
   }
-
+  
   @Override
   public void setPinning(ExtendedBlock b) throws IOException {
     getBlockMap(b).get(b.getLocalBlock()).pinned = true;
   }
-
+  
   @Override
   public boolean getPinning(ExtendedBlock b) throws IOException {
     return getBlockMap(b).get(b.getLocalBlock()).pinned;
   }
-
+  
   @Override
   public boolean isDeletingBlock(String bpid, long blockId) {
     throw new UnsupportedOperationException();
   }
 
+  @Override
+  public ReplicaInfo moveBlockAcrossVolumes(ExtendedBlock block,
+      FsVolumeSpi destination) throws IOException {
+    return null;
+  }
+
   @Override
   public AutoCloseableLock acquireDatasetLock() {
     return datasetLock.acquire();
@@ -1547,3 +1588,4 @@ public class SimulatedFSDataset implements FsDatasetSpi<FsVolumeSpi> {
     return Collections.unmodifiableSet(replicas);
   }
 }
+
