@@ -29,17 +29,19 @@ import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.EnumSet;
+import java.util.HashMap;
 import java.util.HashSet;
 import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
 import java.util.concurrent.atomic.AtomicInteger;
+import java.util.stream.Collectors;
 
 import org.apache.commons.cli.UnrecognizedOptionException;
-import org.apache.commons.lang.math.LongRange;
-import org.apache.commons.logging.Log;
-import org.apache.commons.logging.LogFactory;
+import org.apache.commons.lang3.Range;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
 import org.apache.hadoop.classification.InterfaceAudience.Private;
 import org.apache.hadoop.conf.Configuration;
 import org.apache.hadoop.fs.CommonConfigurationKeysPublic;
@@ -66,8 +68,12 @@ import org.apache.hadoop.yarn.api.protocolrecords.GetApplicationReportRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetApplicationReportResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetApplicationsRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetApplicationsResponse;
+import org.apache.hadoop.yarn.api.protocolrecords.GetAttributesToNodesRequest;
+import org.apache.hadoop.yarn.api.protocolrecords.GetAttributesToNodesResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetClusterMetricsRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetClusterMetricsResponse;
+import org.apache.hadoop.yarn.api.protocolrecords.GetClusterNodeAttributesRequest;
+import org.apache.hadoop.yarn.api.protocolrecords.GetClusterNodeAttributesResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetClusterNodeLabelsRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetClusterNodeLabelsResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetClusterNodesRequest;
@@ -84,6 +90,8 @@ import org.apache.hadoop.yarn.api.protocolrecords.GetNewApplicationRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetNewApplicationResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetNewReservationRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetNewReservationResponse;
+import org.apache.hadoop.yarn.api.protocolrecords.GetNodesToAttributesRequest;
+import org.apache.hadoop.yarn.api.protocolrecords.GetNodesToAttributesResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetNodesToLabelsRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetNodesToLabelsResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetQueueInfoRequest;
@@ -112,8 +120,12 @@ import org.apache.hadoop.yarn.api.protocolrecords.UpdateApplicationPriorityReque
 import org.apache.hadoop.yarn.api.protocolrecords.UpdateApplicationPriorityResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.UpdateApplicationTimeoutsRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.UpdateApplicationTimeoutsResponse;
+import org.apache.hadoop.yarn.api.protocolrecords.GetAllResourceProfilesRequest;
+import org.apache.hadoop.yarn.api.protocolrecords.GetAllResourceProfilesResponse;
 import org.apache.hadoop.yarn.api.protocolrecords.GetAllResourceTypeInfoRequest;
 import org.apache.hadoop.yarn.api.protocolrecords.GetAllResourceTypeInfoResponse;
+import org.apache.hadoop.yarn.api.protocolrecords.GetResourceProfileRequest;
+import org.apache.hadoop.yarn.api.protocolrecords.GetResourceProfileResponse;
 import org.apache.hadoop.yarn.api.records.ApplicationAccessType;
 import org.apache.hadoop.yarn.api.records.ApplicationAttemptId;
 import org.apache.hadoop.yarn.api.records.ApplicationAttemptReport;
@@ -123,8 +135,12 @@ import org.apache.hadoop.yarn.api.records.ApplicationSubmissionContext;
 import org.apache.hadoop.yarn.api.records.ApplicationTimeoutType;
 import org.apache.hadoop.yarn.api.records.ContainerId;
 import org.apache.hadoop.yarn.api.records.ContainerReport;
+import org.apache.hadoop.yarn.api.records.NodeAttribute;
+import org.apache.hadoop.yarn.api.records.NodeAttributeKey;
+import org.apache.hadoop.yarn.api.records.NodeAttributeInfo;
 import org.apache.hadoop.yarn.api.records.NodeReport;
 import org.apache.hadoop.yarn.api.records.NodeState;
+import org.apache.hadoop.yarn.api.records.NodeToAttributeValue;
 import org.apache.hadoop.yarn.api.records.Priority;
 import org.apache.hadoop.yarn.api.records.QueueACL;
 import org.apache.hadoop.yarn.api.records.QueueInfo;
@@ -140,15 +156,19 @@ import org.apache.hadoop.yarn.exceptions.ApplicationAttemptNotFoundException;
 import org.apache.hadoop.yarn.exceptions.ApplicationNotFoundException;
 import org.apache.hadoop.yarn.exceptions.ContainerNotFoundException;
 import org.apache.hadoop.yarn.exceptions.YarnException;
+import org.apache.hadoop.yarn.exceptions.YARNFeatureNotEnabledException;
 import org.apache.hadoop.yarn.factories.RecordFactory;
 import org.apache.hadoop.yarn.factory.providers.RecordFactoryProvider;
 import org.apache.hadoop.yarn.ipc.RPCUtil;
 import org.apache.hadoop.yarn.ipc.YarnRPC;
+import org.apache.hadoop.yarn.nodelabels.AttributeValue;
+import org.apache.hadoop.yarn.nodelabels.NodeAttributesManager;
 import org.apache.hadoop.yarn.security.client.RMDelegationTokenIdentifier;
 import org.apache.hadoop.yarn.server.resourcemanager.RMAuditLogger.AuditConstants;
 import org.apache.hadoop.yarn.server.resourcemanager.RMAuditLogger.Keys;
 import org.apache.hadoop.yarn.server.resourcemanager.nodelabels.RMNodeLabelsManager;
 import org.apache.hadoop.yarn.server.resourcemanager.preprocessor.SubmissionContextPreProcessor;
+import org.apache.hadoop.yarn.server.resourcemanager.resource.ResourceProfilesManager;
 import org.apache.hadoop.yarn.server.resourcemanager.reservation.Plan;
 import org.apache.hadoop.yarn.server.resourcemanager.reservation.ReservationAllocation;
 import org.apache.hadoop.yarn.server.resourcemanager.reservation.ReservationInputValidator;
@@ -177,11 +197,11 @@ import org.apache.hadoop.yarn.server.utils.BuilderUtils;
 import org.apache.hadoop.yarn.util.Clock;
 import org.apache.hadoop.yarn.util.Records;
 import org.apache.hadoop.yarn.util.UTCClock;
+
+import com.google.common.annotations.VisibleForTesting;
 import org.apache.hadoop.yarn.util.resource.ResourceUtils;
 import org.apache.hadoop.yarn.util.timeline.TimelineUtils;
 
-import com.google.common.annotations.VisibleForTesting;
-import com.google.common.base.CharMatcher;
 
 /**
  * The client interface to the Resource Manager. This module handles all the rpc
@@ -191,7 +211,8 @@ public class ClientRMService extends AbstractService implements
     ApplicationClientProtocol {
   private static final ArrayList<ApplicationReport> EMPTY_APPS_REPORT = new ArrayList<ApplicationReport>();
 
-  private static final Log LOG = LogFactory.getLog(ClientRMService.class);
+  private static final Logger LOG =
+      LoggerFactory.getLogger(ClientRMService.class);
 
   final private AtomicInteger applicationCounter = new AtomicInteger(0);
   final private YarnScheduler scheduler;
@@ -214,11 +235,14 @@ public class ClientRMService extends AbstractService implements
 
   private SubmissionContextPreProcessor contextPreProcessor;
 
-  private boolean displayPerUserApps = false;
+  private boolean filterAppsByUser = false;
 
   private static final EnumSet<RMAppState> ACTIVE_APP_STATES = EnumSet.of(
       RMAppState.ACCEPTED, RMAppState.RUNNING);
 
+  private ResourceProfilesManager resourceProfilesManager;
+  private boolean timelineServiceV2Enabled;
+
   public ClientRMService(RMContext rmContext, YarnScheduler scheduler,
       RMAppManager rmAppManager, ApplicationACLsManager applicationACLsManager,
       QueueACLsManager queueACLsManager,
@@ -241,6 +265,7 @@ public class ClientRMService extends AbstractService implements
     this.reservationSystem = rmContext.getReservationSystem();
     this.clock = clock;
     this.rValidator = new ReservationInputValidator(clock);
+    resourceProfilesManager = rmContext.getResourceProfilesManager();
   }
 
   @Override
@@ -260,6 +285,11 @@ public class ClientRMService extends AbstractService implements
             conf.getInt(YarnConfiguration.RM_CLIENT_THREAD_COUNT, 
                 YarnConfiguration.DEFAULT_RM_CLIENT_THREAD_COUNT));
     
+    this.server.addTerseExceptions(ApplicationNotFoundException.class,
+        ApplicationAttemptNotFoundException.class,
+        ContainerNotFoundException.class,
+        YARNFeatureNotEnabledException.class);
+
     // Enable service authorization?
     if (conf.getBoolean(
         CommonConfigurationKeysPublic.HADOOP_SECURITY_AUTHORIZATION, 
@@ -274,8 +304,8 @@ public class ClientRMService extends AbstractService implements
       refreshServiceAcls(conf, RMPolicyProvider.getInstance());
     }
 
-    this.displayPerUserApps  = conf.getBoolean(
-        YarnConfiguration.DISPLAY_APPS_FOR_LOGGED_IN_USER,
+    this.filterAppsByUser  = conf.getBoolean(
+        YarnConfiguration.FILTER_ENTITY_LIST_BY_USER,
         YarnConfiguration.DEFAULT_DISPLAY_APPS_FOR_LOGGED_IN_USER);
 
     this.server.start();
@@ -283,6 +313,9 @@ public class ClientRMService extends AbstractService implements
                                                YarnConfiguration.RM_ADDRESS,
                                                YarnConfiguration.DEFAULT_RM_ADDRESS,
                                                server.getListenerAddress());
+    this.timelineServiceV2Enabled = YarnConfiguration.
+        timelineServiceV2Enabled(conf);
+
     if (conf.getBoolean(
         YarnConfiguration.RM_SUBMISSION_PREPROCESSOR_ENABLED,
         YarnConfiguration.DEFAULT_RM_SUBMISSION_PREPROCESSOR_ENABLED)) {
@@ -407,14 +440,13 @@ public class ClientRMService extends AbstractService implements
       GetApplicationAttemptReportRequest request) throws YarnException,
       IOException {
     ApplicationId applicationId
-      = request.getApplicationAttemptId().getApplicationId();
+        = request.getApplicationAttemptId().getApplicationId();
     ApplicationAttemptId appAttemptId = request.getApplicationAttemptId();
-
     UserGroupInformation callerUGI = getCallerUgi(applicationId,
-      AuditConstants.GET_APP_ATTEMPT_REPORT);
+        AuditConstants.GET_APP_ATTEMPT_REPORT);
     RMApp application = verifyUserAccessForRMApp(applicationId, callerUGI,
-       AuditConstants.GET_APP_ATTEMPT_REPORT, ApplicationAccessType.VIEW_APP,
-       false);
+        AuditConstants.GET_APP_ATTEMPT_REPORT, ApplicationAccessType.VIEW_APP,
+        false);
 
     boolean allowAccess = checkAccess(callerUGI, application.getUser(),
         ApplicationAccessType.VIEW_APP, application);
@@ -440,13 +472,11 @@ public class ClientRMService extends AbstractService implements
   public GetApplicationAttemptsResponse getApplicationAttempts(
       GetApplicationAttemptsRequest request) throws YarnException, IOException {
     ApplicationId appId = request.getApplicationId();
-
     UserGroupInformation callerUGI = getCallerUgi(appId,
-      AuditConstants.GET_APP_ATTEMPTS);
+        AuditConstants.GET_APP_ATTEMPTS);
     RMApp application = verifyUserAccessForRMApp(appId, callerUGI,
-      AuditConstants.GET_APP_ATTEMPTS, ApplicationAccessType.VIEW_APP,
-      false);
-
+        AuditConstants.GET_APP_ATTEMPTS, ApplicationAccessType.VIEW_APP,
+        false);
     boolean allowAccess = checkAccess(callerUGI, application.getUser(),
         ApplicationAccessType.VIEW_APP, application);
     GetApplicationAttemptsResponse response = null;
@@ -484,13 +514,11 @@ public class ClientRMService extends AbstractService implements
     }
     ApplicationAttemptId appAttemptId = containerId.getApplicationAttemptId();
     ApplicationId appId = appAttemptId.getApplicationId();
-
     UserGroupInformation callerUGI = getCallerUgi(appId,
         AuditConstants.GET_CONTAINER_REPORT);
     RMApp application = verifyUserAccessForRMApp(appId, callerUGI,
         AuditConstants.GET_CONTAINER_REPORT, ApplicationAccessType.VIEW_APP,
         false);
-
     boolean allowAccess = checkAccess(callerUGI, application.getUser(),
         ApplicationAccessType.VIEW_APP, application);
     GetContainerReportResponse response = null;
@@ -527,13 +555,10 @@ public class ClientRMService extends AbstractService implements
       throws YarnException, IOException {
     ApplicationAttemptId appAttemptId = request.getApplicationAttemptId();
     ApplicationId appId = appAttemptId.getApplicationId();
-
     UserGroupInformation callerUGI = getCallerUgi(appId,
         AuditConstants.GET_CONTAINERS);
     RMApp application = verifyUserAccessForRMApp(appId, callerUGI,
-        AuditConstants.GET_CONTAINERS, ApplicationAccessType.VIEW_APP,
-        false);
-
+        AuditConstants.GET_CONTAINERS, ApplicationAccessType.VIEW_APP, false);
     boolean allowAccess = checkAccess(callerUGI, application.getUser(),
         ApplicationAccessType.VIEW_APP, application);
     GetContainersResponse response = null;
@@ -575,12 +600,10 @@ public class ClientRMService extends AbstractService implements
     // checked here, those that are dependent on RM configuration are validated
     // in RMAppManager.
 
-    UserGroupInformation userUgi = null;
     String user = null;
     try {
       // Safety
-      userUgi = UserGroupInformation.getCurrentUser();
-      user = userUgi.getShortUserName();
+      user = UserGroupInformation.getCurrentUser().getShortUserName();
     } catch (IOException ie) {
       LOG.warn("Unable to get the current user.", ie);
       RMAuditLogger.logFailure(user, AuditConstants.SUBMIT_APP_REQUEST,
@@ -592,7 +615,7 @@ public class ClientRMService extends AbstractService implements
 
     checkTags(submissionContext.getApplicationTags());
 
-    if (YarnConfiguration.timelineServiceV2Enabled(getConfig())) {
+    if (timelineServiceV2Enabled) {
       // Sanity check for flow run
       String value = null;
       try {
@@ -602,6 +625,7 @@ public class ClientRMService extends AbstractService implements
                   TimelineUtils.FLOW_RUN_ID_TAG_PREFIX.toLowerCase() + ":")) {
             value = tag.substring(TimelineUtils.FLOW_RUN_ID_TAG_PREFIX.length()
                 + 1);
+            // In order to check the number format
             Long.valueOf(value);
           }
         }
@@ -670,7 +694,7 @@ public class ClientRMService extends AbstractService implements
     try {
       // call RMAppManager to submit application directly
       rmAppManager.submitApplication(submissionContext,
-          System.currentTimeMillis(), userUgi);
+          System.currentTimeMillis(), user);
 
       LOG.info("Application with id " + applicationId.getId() + 
           " submitted by user " + user);
@@ -704,7 +728,7 @@ public class ClientRMService extends AbstractService implements
         AuditConstants.FAIL_ATTEMPT_REQUEST);
     RMApp application = verifyUserAccessForRMApp(applicationId, callerUGI,
         AuditConstants.FAIL_ATTEMPT_REQUEST, ApplicationAccessType.MODIFY_APP,
-       true);
+        true);
 
     RMAppAttempt appAttempt = application.getAppAttempts().get(attemptId);
     if (appAttempt == null) {
@@ -717,7 +741,7 @@ public class ClientRMService extends AbstractService implements
 
     if (application.isAppInCompletedStates()) {
       RMAuditLogger.logSuccess(callerUGI.getShortUserName(),
-      AuditConstants.FAIL_ATTEMPT_REQUEST, "ClientRMService",
+          AuditConstants.FAIL_ATTEMPT_REQUEST, "ClientRMService",
           applicationId);
       return response;
     }
@@ -751,7 +775,7 @@ public class ClientRMService extends AbstractService implements
             new IllegalArgumentException("Tag " + tag + " is too long, "
                 + "maximum allowed length of a tag is " + appMaxTagLength));
       }
-      if (!CharMatcher.ASCII.matchesAllOf(tag)) {
+      if (!org.apache.commons.lang3.StringUtils.isAsciiPrintable(tag)) {
         throw RPCUtil.getRemoteException(new IllegalArgumentException(
             "A tag can only have ASCII " + "characters! Invalid tag - " + tag));
       }
@@ -772,11 +796,10 @@ public class ClientRMService extends AbstractService implements
     } catch (IOException ie) {
       LOG.info("Error getting UGI ", ie);
       RMAuditLogger.logFailure("UNKNOWN", AuditConstants.KILL_APP_REQUEST,
-          "UNKNOWN", "ClientRMService" , "Error getting UGI",
-          applicationId, callerContext);
+              "UNKNOWN", "ClientRMService", "Error getting UGI",
+              applicationId, callerContext);
       throw RPCUtil.getRemoteException(ie);
     }
-
     RMApp application = this.rmContext.getRMApps().get(applicationId);
     if (application == null) {
       RMAuditLogger.logFailure(callerUGI.getUserName(),
@@ -811,7 +834,7 @@ public class ClientRMService extends AbstractService implements
       message.append(" at ").append(remoteAddress.getHostAddress());
     }
 
-    String diagnostics = org.apache.commons.lang.StringUtils
+    String diagnostics = org.apache.commons.lang3.StringUtils
         .trimToNull(request.getDiagnostics());
     if (diagnostics != null) {
       message.append(" with diagnostic message: ");
@@ -869,8 +892,8 @@ public class ClientRMService extends AbstractService implements
     Set<String> queues = request.getQueues();
     Set<String> tags = request.getApplicationTags();
     long limit = request.getLimit();
-    LongRange start = request.getStartRange();
-    LongRange finish = request.getFinishRange();
+    Range<Long> start = request.getStartRange();
+    Range<Long> finish = request.getFinishRange();
     ApplicationsRequestScope scope = request.getScope();
 
     final Map<ApplicationId, RMApp> apps = rmContext.getRMApps();
@@ -945,11 +968,11 @@ public class ClientRMService extends AbstractService implements
         continue;
       }
 
-      if (start != null && !start.containsLong(application.getStartTime())) {
+      if (start != null && !start.contains(application.getStartTime())) {
         continue;
       }
 
-      if (finish != null && !finish.containsLong(application.getFinishTime())) {
+      if (finish != null && !finish.contains(application.getFinishTime())) {
         continue;
       }
 
@@ -979,7 +1002,7 @@ public class ClientRMService extends AbstractService implements
 
       // Given RM is configured to display apps per user, skip apps to which
       // this caller doesn't have access to view.
-      if (displayPerUserApps && !allowAccess) {
+      if (filterAppsByUser && !allowAccess) {
         continue;
       }
 
@@ -1085,7 +1108,7 @@ public class ClientRMService extends AbstractService implements
     return response;
   }
 
-  private NodeReport createNodeReports(RMNode rmNode) {    
+  private NodeReport createNodeReports(RMNode rmNode) {
     SchedulerNodeReport schedulerNodeReport = 
         scheduler.getNodeReport(rmNode.getNodeID());
     Resource used = BuilderUtils.newResource(0, 0);
@@ -1093,8 +1116,9 @@ public class ClientRMService extends AbstractService implements
     if (schedulerNodeReport != null) {
       used = schedulerNodeReport.getUsedResource();
       numContainers = schedulerNodeReport.getNumContainers();
-    } 
+    }
 
+    Set<NodeAttribute> attrs = rmNode.getAllNodeAttributes();
     NodeReport report =
         BuilderUtils.newNodeReport(rmNode.getNodeID(), rmNode.getState(),
             rmNode.getHttpAddress(), rmNode.getRackName(), used,
@@ -1102,7 +1126,7 @@ public class ClientRMService extends AbstractService implements
             rmNode.getHealthReport(), rmNode.getLastHealthReportTime(),
             rmNode.getNodeLabels(), rmNode.getAggregatedContainersUtilization(),
             rmNode.getNodeUtilization(), rmNode.getDecommissioningTimeout(),
-            null);
+            null, attrs);
 
     return report;
   }
@@ -1210,20 +1234,34 @@ public class ClientRMService extends AbstractService implements
     UserGroupInformation callerUGI = getCallerUgi(applicationId,
         AuditConstants.MOVE_APP_REQUEST);
     RMApp application = verifyUserAccessForRMApp(applicationId, callerUGI,
-         AuditConstants.MOVE_APP_REQUEST, ApplicationAccessType.MODIFY_APP,
+        AuditConstants.MOVE_APP_REQUEST, ApplicationAccessType.MODIFY_APP,
         true);
 
+    String targetQueue = request.getTargetQueue();
+    if (!accessToTargetQueueAllowed(callerUGI, application, targetQueue)) {
+      RMAuditLogger.logFailure(callerUGI.getShortUserName(),
+          AuditConstants.MOVE_APP_REQUEST, "Target queue doesn't exist or user"
+              + " doesn't have permissions to submit to target queue: "
+              + targetQueue, "ClientRMService",
+          AuditConstants.UNAUTHORIZED_USER, applicationId);
+      throw RPCUtil.getRemoteException(new AccessControlException("User "
+          + callerUGI.getShortUserName() + " cannot submit applications to"
+          + " target queue or the target queue doesn't exist: "
+          + targetQueue + " while moving " + applicationId));
+    }
+
     // Moves only allowed when app is in a state that means it is tracked by
     // the scheduler. Introducing SUBMITTED state also to this list as there
     // could be a corner scenario that app may not be in Scheduler in SUBMITTED
     // state.
     if (!ACTIVE_APP_STATES.contains(application.getState())) {
-      String msg = "App in " + application.getState() + " state cannot be moved.";
+      String msg = "App in " + application.getState() +
+          " state cannot be moved.";
       RMAuditLogger.logFailure(callerUGI.getShortUserName(),
           AuditConstants.MOVE_APP_REQUEST, "UNKNOWN", "ClientRMService", msg);
       throw new YarnException(msg);
     }
-    
+
     try {
       this.rmAppManager.moveApplicationAcrossQueue(
           application.getApplicationId(),
@@ -1241,6 +1279,24 @@ public class ClientRMService extends AbstractService implements
         .newRecordInstance(MoveApplicationAcrossQueuesResponse.class);
   }
 
+  /**
+   * Check if the submission of an application to the target queue is allowed.
+   * @param callerUGI the caller UGI
+   * @param application the application to move
+   * @param targetQueue the queue to move the application to
+   * @return true if submission is allowed, false otherwise
+   */
+  private boolean accessToTargetQueueAllowed(UserGroupInformation callerUGI,
+      RMApp application, String targetQueue) {
+    return
+        queueACLsManager.checkAccess(callerUGI,
+            QueueACL.SUBMIT_APPLICATIONS, application,
+            Server.getRemoteAddress(), null, targetQueue) ||
+        queueACLsManager.checkAccess(callerUGI,
+            QueueACL.ADMINISTER_QUEUE, application,
+            Server.getRemoteAddress(), null, targetQueue);
+  }
+
   private String getRenewerForToken(Token<RMDelegationTokenIdentifier> token)
       throws IOException {
     UserGroupInformation user = UserGroupInformation.getCurrentUser();
@@ -1468,9 +1524,7 @@ public class ClientRMService extends AbstractService implements
   public GetNodesToLabelsResponse getNodeToLabels(
       GetNodesToLabelsRequest request) throws YarnException, IOException {
     RMNodeLabelsManager labelsMgr = rmContext.getNodeLabelManager();
-    GetNodesToLabelsResponse response =
-        GetNodesToLabelsResponse.newInstance(labelsMgr.getNodeLabels());
-    return response;
+    return GetNodesToLabelsResponse.newInstance(labelsMgr.getNodeLabels());
   }
 
   @Override
@@ -1478,8 +1532,7 @@ public class ClientRMService extends AbstractService implements
       GetLabelsToNodesRequest request) throws YarnException, IOException {
     RMNodeLabelsManager labelsMgr = rmContext.getNodeLabelManager();
     if (request.getNodeLabels() == null || request.getNodeLabels().isEmpty()) {
-      return GetLabelsToNodesResponse.newInstance(
-          labelsMgr.getLabelsToNodes());
+      return GetLabelsToNodesResponse.newInstance(labelsMgr.getLabelsToNodes());
     } else {
       return GetLabelsToNodesResponse.newInstance(
           labelsMgr.getLabelsToNodes(request.getNodeLabels()));
@@ -1490,10 +1543,8 @@ public class ClientRMService extends AbstractService implements
   public GetClusterNodeLabelsResponse getClusterNodeLabels(
       GetClusterNodeLabelsRequest request) throws YarnException, IOException {
     RMNodeLabelsManager labelsMgr = rmContext.getNodeLabelManager();
-    GetClusterNodeLabelsResponse response =
-        GetClusterNodeLabelsResponse.newInstance(
+    return GetClusterNodeLabelsResponse.newInstance(
             labelsMgr.getClusterNodeLabels());
-    return response;
   }
 
   private void checkReservationSystem()
@@ -1807,17 +1858,17 @@ public class ClientRMService extends AbstractService implements
   }
 
   private RMApp verifyUserAccessForRMApp(ApplicationId applicationId,
-                                         UserGroupInformation callerUGI, String operation,
-                                         ApplicationAccessType accessType,
-                                         boolean needCheckAccess) throws YarnException {
+      UserGroupInformation callerUGI, String operation,
+      ApplicationAccessType accessType,
+      boolean needCheckAccess) throws YarnException {
     RMApp application = this.rmContext.getRMApps().get(applicationId);
     if (application == null) {
       RMAuditLogger.logFailure(callerUGI.getUserName(), operation, "UNKNOWN",
-              "ClientRMService",
-              "Trying to " + operation + " of an absent application",
-              applicationId);
-      // If the RM doesn't have the application, throw
-      // ApplicationNotFoundException and let client to handle.
+          "ClientRMService",
+          "Trying to " + operation + " of an absent application",
+          applicationId);
+        // If the RM doesn't have the application, throw
+        // ApplicationNotFoundException and let client to handle.
       throw new ApplicationNotFoundException("Application with id '"
               + applicationId + "' doesn't exist in RM. "
               + "Please check that the job "
@@ -1840,9 +1891,23 @@ public class ClientRMService extends AbstractService implements
     return application;
   }
 
-  @VisibleForTesting
-  public void setDisplayPerUserApps(boolean displayPerUserApps) {
-    this.displayPerUserApps = displayPerUserApps;
+  @Override
+  public GetAllResourceProfilesResponse getResourceProfiles(
+      GetAllResourceProfilesRequest request) throws YarnException, IOException {
+    GetAllResourceProfilesResponse response =
+        GetAllResourceProfilesResponse.newInstance();
+    response.setResourceProfiles(resourceProfilesManager.getResourceProfiles());
+    return response;
+  }
+
+  @Override
+  public GetResourceProfileResponse getResourceProfile(
+      GetResourceProfileRequest request) throws YarnException, IOException {
+    GetResourceProfileResponse response =
+        GetResourceProfileResponse.newInstance();
+    response.setResource(
+        resourceProfilesManager.getProfile(request.getProfileName()));
+    return response;
   }
 
   @Override
@@ -1853,4 +1918,61 @@ public class ClientRMService extends AbstractService implements
     response.setResourceTypeInfo(ResourceUtils.getResourcesTypeInfo());
     return response;
   }
+
+  @Override
+  public GetAttributesToNodesResponse getAttributesToNodes(
+      GetAttributesToNodesRequest request) throws YarnException, IOException {
+    NodeAttributesManager attributesManager =
+        rmContext.getNodeAttributesManager();
+    Map<NodeAttributeKey, List<NodeToAttributeValue>> attrToNodesWithStrVal =
+        new HashMap<>();
+    Map<NodeAttributeKey, Map<String, AttributeValue>> attributesToNodes =
+        attributesManager.getAttributesToNodes(request.getNodeAttributes());
+    for (Map.Entry<NodeAttributeKey, Map<String, AttributeValue>> attrib :
+          attributesToNodes.entrySet()) {
+      Map<String, AttributeValue> nodesToVal = attrib.getValue();
+      List<NodeToAttributeValue> nodeToAttrValList = new ArrayList<>();
+      for (Map.Entry<String, AttributeValue> nodeToVal : nodesToVal
+          .entrySet()) {
+        nodeToAttrValList.add(NodeToAttributeValue
+            .newInstance(nodeToVal.getKey(), nodeToVal.getValue().getValue()));
+      }
+      attrToNodesWithStrVal.put(attrib.getKey(), nodeToAttrValList);
+    }
+    GetAttributesToNodesResponse response =
+        GetAttributesToNodesResponse.newInstance(attrToNodesWithStrVal);
+    return response;
+  }
+
+  @Override
+  public GetClusterNodeAttributesResponse getClusterNodeAttributes(
+      GetClusterNodeAttributesRequest request)
+      throws YarnException, IOException {
+    NodeAttributesManager attributesManager =
+        rmContext.getNodeAttributesManager();
+    Set<NodeAttribute> attributes =
+        attributesManager.getClusterNodeAttributes(null);
+
+    GetClusterNodeAttributesResponse response =
+        GetClusterNodeAttributesResponse.newInstance(
+            attributes.stream().map(attr -> NodeAttributeInfo.newInstance(attr))
+                .collect(Collectors.toSet()));
+    return response;
+  }
+
+  @Override
+  public GetNodesToAttributesResponse getNodesToAttributes(
+      GetNodesToAttributesRequest request) throws YarnException, IOException {
+    NodeAttributesManager attributesManager =
+        rmContext.getNodeAttributesManager();
+    GetNodesToAttributesResponse response = GetNodesToAttributesResponse
+        .newInstance(
+            attributesManager.getNodesToAttributes(request.getHostNames()));
+    return response;
+  }
+
+  @VisibleForTesting
+  public void setDisplayPerUserApps(boolean displayPerUserApps) {
+    this.filterAppsByUser = displayPerUserApps;
+  }
 }
