@@ -22,8 +22,6 @@ import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertNotNull;
 import static org.junit.Assert.assertTrue;
-import static org.mockito.Mockito.mock;
-import static org.mockito.Mockito.when;
 
 import java.io.ByteArrayOutputStream;
 import java.io.IOException;
@@ -34,14 +32,10 @@ import java.net.HttpCookie;
 import java.net.HttpURLConnection;
 import java.net.URI;
 import java.net.URL;
-import java.net.SocketTimeoutException;
-import java.util.Collections;
 import java.util.Enumeration;
 import java.util.List;
 import java.util.Map;
 
-import javax.servlet.ServletConfig;
-import javax.servlet.ServletContext;
 import javax.servlet.ServletException;
 import javax.servlet.http.HttpServlet;
 import javax.servlet.http.HttpServletRequest;
@@ -61,12 +55,14 @@ import org.apache.hadoop.yarn.conf.YarnConfiguration;
 import org.apache.hadoop.yarn.exceptions.ApplicationNotFoundException;
 import org.apache.hadoop.yarn.exceptions.YarnException;
 import org.apache.hadoop.yarn.webapp.util.WebAppUtils;
+import org.eclipse.jetty.server.ServerConnector;
+import org.eclipse.jetty.servlet.ServletContextHandler;
+import org.eclipse.jetty.util.thread.QueuedThreadPool;
 import org.junit.AfterClass;
 import org.junit.BeforeClass;
 import org.junit.Test;
-import org.mortbay.jetty.Server;
-import org.mortbay.jetty.servlet.Context;
-import org.mortbay.jetty.servlet.ServletHolder;
+import org.eclipse.jetty.server.Server;
+import org.eclipse.jetty.servlet.ServletHolder;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -92,14 +88,14 @@ public class TestWebAppProxyServlet {
   @BeforeClass
   public static void start() throws Exception {
     server = new Server(0);
-    Context context = new Context();
+    ((QueuedThreadPool)server.getThreadPool()).setMaxThreads(20);
+    ServletContextHandler context = new ServletContextHandler();
     context.setContextPath("/foo");
     server.setHandler(context);
     context.addServlet(new ServletHolder(TestServlet.class), "/bar");
-    context.addServlet(new ServletHolder(TimeOutTestServlet.class), "/timeout");
-    server.getConnectors()[0].setHost("localhost");
+    ((ServerConnector)server.getConnectors()[0]).setHost("localhost");
     server.start();
-    originalPort = server.getConnectors()[0].getLocalPort();
+    originalPort = ((ServerConnector)server.getConnectors()[0]).getLocalPort();
     LOG.info("Running embedded servlet container at: http://localhost:"
         + originalPort);
     // This property needs to be set otherwise CORS Headers will be dropped
@@ -144,38 +140,15 @@ public class TestWebAppProxyServlet {
     }
   }
 
-  @SuppressWarnings("serial")
-  public static class TimeOutTestServlet extends HttpServlet {
-
-    @Override
-    protected void doGet(HttpServletRequest req, HttpServletResponse resp)
-        throws ServletException, IOException {
-      try {
-        Thread.sleep(10 * 1000);
-      } catch (InterruptedException e) {
-        LOG.warn("doGet() interrupted", e);
-        resp.setStatus(HttpServletResponse.SC_BAD_REQUEST);
-        return;
-      }
-      resp.setStatus(HttpServletResponse.SC_OK);
-    }
-
-    @Override
-    protected void doPost(HttpServletRequest req, HttpServletResponse resp)
-        throws ServletException, IOException {
-      resp.setStatus(HttpServletResponse.SC_OK);
-    }
-  }
-
   @Test(timeout=5000)
   public void testWebAppProxyServlet() throws Exception {
     configuration.set(YarnConfiguration.PROXY_ADDRESS, "localhost:9090");
-    // overriding num of web server threads, see HttpServer.HTTP_MAXTHREADS 
-    configuration.setInt("hadoop.http.max.threads", 5);
+    // overriding num of web server threads, see HttpServer.HTTP_MAXTHREADS
+    configuration.setInt("hadoop.http.max.threads", 10);
     WebAppProxyServerForTest proxy = new WebAppProxyServerForTest();
     proxy.init(configuration);
     proxy.start();
-    
+
     int proxyPort = proxy.proxy.proxyServer.getConnectorAddress(0).getPort();
     AppReportFetcherForTest appReportFetcher = proxy.proxy.appReportFetcher;
 
@@ -278,50 +251,11 @@ public class TestWebAppProxyServlet {
     }
   }
 
-  @Test(expected = SocketTimeoutException.class)
-  public void testWebAppProxyConnectionTimeout()
-      throws IOException, ServletException{
-    HttpServletRequest request = mock(HttpServletRequest.class);
-    when(request.getMethod()).thenReturn("GET");
-    when(request.getRemoteUser()).thenReturn("dr.who");
-    when(request.getPathInfo()).thenReturn("/application_00_0");
-    when(request.getHeaderNames()).thenReturn(Collections.emptyEnumeration());
-
-    HttpServletResponse response = mock(HttpServletResponse.class);
-    when(response.getOutputStream()).thenReturn(null);
-
-    WebAppProxyServlet servlet = new WebAppProxyServlet();
-    YarnConfiguration conf = new YarnConfiguration();
-    conf.setBoolean(YarnConfiguration.RM_PROXY_TIMEOUT_ENABLED,
-        true);
-    conf.setInt(YarnConfiguration.RM_PROXY_CONNECTION_TIMEOUT,
-        1000);
-
-    servlet.setConf(conf);
-
-    ServletConfig config = mock(ServletConfig.class);
-    ServletContext context = mock(ServletContext.class);
-    when(config.getServletContext()).thenReturn(context);
-
-    AppReportFetcherForTest appReportFetcher =
-        new AppReportFetcherForTest(new YarnConfiguration());
-
-    when(config.getServletContext()
-        .getAttribute(WebAppProxy.FETCHER_ATTRIBUTE))
-        .thenReturn(appReportFetcher);
-
-    appReportFetcher.answer = 7;
-
-    servlet.init(config);
-    servlet.doGet(request, response);
-
-  }
-
   @Test(timeout=5000)
   public void testAppReportForEmptyTrackingUrl() throws Exception {
     configuration.set(YarnConfiguration.PROXY_ADDRESS, "localhost:9090");
     // overriding num of web server threads, see HttpServer.HTTP_MAXTHREADS
-    configuration.setInt("hadoop.http.max.threads", 5);
+    configuration.setInt("hadoop.http.max.threads", 10);
     WebAppProxyServerForTest proxy = new WebAppProxyServerForTest();
     proxy.init(configuration);
     proxy.start();
@@ -379,7 +313,7 @@ public class TestWebAppProxyServlet {
   public void testWebAppProxyPassThroughHeaders() throws Exception {
     Configuration configuration = new Configuration();
     configuration.set(YarnConfiguration.PROXY_ADDRESS, "localhost:9091");
-    configuration.setInt("hadoop.http.max.threads", 5);
+    configuration.setInt("hadoop.http.max.threads", 10);
     WebAppProxyServerForTest proxy = new WebAppProxyServerForTest();
     proxy.init(configuration);
     proxy.start();
@@ -399,13 +333,12 @@ public class TestWebAppProxyServlet {
       assertEquals(proxyConn.getRequestProperties().size(), 4);
       proxyConn.connect();
       assertEquals(HttpURLConnection.HTTP_OK, proxyConn.getResponseCode());
-      // Verify if number of headers received by end server is 9.
-      // This should match WebAppProxyServlet#PASS_THROUGH_HEADERS.
-      // Nine headers include Accept, Host, Connection, User-Agent, Cookie,
-      // Origin, Access-Control-Request-Method, Accept-Encoding, and
+      // Verify if number of headers received by end server is 8.
+      // Eight headers include Accept, Host, Connection, User-Agent, Cookie,
+      // Origin, Access-Control-Request-Method and
       // Access-Control-Request-Headers. Pls note that Unknown-Header is dropped
       // by proxy as it is not in the list of allowed headers.
-      assertEquals(numberOfHeaders, 9);
+      assertEquals(numberOfHeaders, 8);
       assertFalse(hasUnknownHeader);
     } finally {
       proxy.close();
@@ -450,47 +383,6 @@ public class TestWebAppProxyServlet {
     }
   }
 
-  /**
-   * Test header injections are not done.
-   */
-  @Test(timeout=5000)
-  public void testWebAppProxyServerHeaderInjection() throws Exception {
-    WebAppProxyServer mainServer = null;
-    Configuration conf = new YarnConfiguration();
-    conf.set(YarnConfiguration.PROXY_ADDRESS, "localhost:9099");
-    try {
-      mainServer = WebAppProxyServer.startServer(conf);
-      int counter = 20;
-
-      URL wrongUrl = new URL(
-          "http://localhost:9099/proxy/%C4%8D%C4%8ASomeCustomInjectedHeader:%20"
-          + "injected_headerVal_1484290871375_0113/");
-      HttpURLConnection proxyConn = null;
-      while (counter > 0) {
-        counter--;
-        try {
-          proxyConn = (HttpURLConnection) wrongUrl.openConnection();
-          proxyConn.connect();
-          proxyConn.getResponseCode();
-          // server started ok
-          counter = 0;
-        } catch (Exception e) {
-          Thread.sleep(100);
-        }
-      }
-      assertNotNull(proxyConn);
-      // wrong application Id
-      assertEquals(HttpURLConnection.HTTP_INTERNAL_ERROR,
-          proxyConn.getResponseCode());
-      assertTrue("Header injection happened",
-          proxyConn.getHeaderField("SomeCustomInjectedHeader") == null);
-    } finally {
-      if (mainServer != null) {
-        mainServer.stop();
-      }
-    }
-  }
-
   private String readInputStream(InputStream input) throws Exception {
     ByteArrayOutputStream data = new ByteArrayOutputStream();
     byte[] buffer = new byte[512];
@@ -501,14 +393,14 @@ public class TestWebAppProxyServlet {
     return new String(data.toByteArray(), "UTF-8");
   }
 
-  private boolean isResponseCookiePresent(HttpURLConnection proxyConn, 
+  private boolean isResponseCookiePresent(HttpURLConnection proxyConn,
       String expectedName, String expectedValue) {
     Map<String, List<String>> headerFields = proxyConn.getHeaderFields();
     List<String> cookiesHeader = headerFields.get("Set-Cookie");
     if (cookiesHeader != null) {
       for (String cookie : cookiesHeader) {
         HttpCookie c = HttpCookie.parse(cookie).get(0);
-        if (c.getName().equals(expectedName) 
+        if (c.getName().equals(expectedName)
             && c.getValue().equals(expectedValue)) {
           return true;
         }
@@ -548,7 +440,7 @@ public class TestWebAppProxyServlet {
   }
 
   private class WebAppProxyForTest extends WebAppProxy {
-    
+
     HttpServer2 proxyServer;
     AppReportFetcherForTest appReportFetcher;
 
@@ -558,7 +450,7 @@ public class TestWebAppProxyServlet {
       String bindAddress = conf.get(YarnConfiguration.PROXY_ADDRESS);
       bindAddress = StringUtils.split(bindAddress, ':')[0];
       AccessControlList acl = new AccessControlList(
-          conf.get(YarnConfiguration.YARN_ADMIN_ACL, 
+          conf.get(YarnConfiguration.YARN_ADMIN_ACL,
           YarnConfiguration.DEFAULT_YARN_ADMIN_ACL));
       proxyServer = new HttpServer2.Builder()
           .setName("proxy")
@@ -575,11 +467,11 @@ public class TestWebAppProxyServlet {
       proxyServer.setAttribute(FETCHER_ATTRIBUTE,
           appReportFetcher );
       proxyServer.setAttribute(IS_SECURITY_ENABLED_ATTRIBUTE, Boolean.TRUE);
-      
+
       String proxy = WebAppUtils.getProxyHostAndPort(conf);
       String[] proxyParts = proxy.split(":");
       String proxyHost = proxyParts[0];
-      
+
       proxyServer.setAttribute(PROXY_HOST_ATTRIBUTE, proxyHost);
       proxyServer.start();
       LOG.info("Proxy server is started at port {}",
@@ -590,7 +482,7 @@ public class TestWebAppProxyServlet {
 
   private class AppReportFetcherForTest extends AppReportFetcher {
     int answer = 0;
-    
+
     public AppReportFetcherForTest(Configuration conf) {
       super(conf);
     }
@@ -623,12 +515,6 @@ public class TestWebAppProxyServlet {
         return result;
       } else if (answer == 6) {
         return getDefaultApplicationReport(appId, false);
-      } else if (answer == 7) {
-        // test connection timeout
-        FetchedAppReport result = getDefaultApplicationReport(appId);
-        result.getApplicationReport().setOriginalTrackingUrl("localhost:"
-            + originalPort + "/foo/timeout?a=b#main");
-        return result;
       }
       return null;
     }
